/*
 * mygame.cpp: ɮxCclassimplementation
 * Copyright (C) 2002 Woei-Kae Chen <wkc@csie.ntut.edu.tw>
 *
 * This file is part of game, a free game development platform for windows.
 *
 * game is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * game is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 *		91-02-23 V3.0
 *			Add CAnimation Class.
 *      91-03-04 V3.1
 *          Add codes to demostrate the use of CMovingBitmap::ShowBitmap(CMovingBitmap &).
 */

#include "stdafx.h"
#include "Resource.h"
#include <mmsystem.h>
#include <ddraw.h>
#include <dsound.h>
#include <dmusici.h>
#include "gamelib.h"
#include "mygame.h"

/////////////////////////////////////////////////////////////////////////////
// CBall: Ball class
// oclassѲyBʻPIB
// noIs(B)UدOAOiHUC{ON
/////////////////////////////////////////////////////////////////////////////

CBall::CBall()
{
	SetDiffX(0);
	SetDiffY(0);
}

void CBall::SetDiffX(int dx)
{
	diff.x = dx;
}

void CBall::SetDiffY(int dy)
{
	diff.y = dy;
}

int  CBall::GetDiffX()
{
	return diff.x;
}

int  CBall::GetDiffY()
{
	return diff.y;
}

void CBall::Move()
{
	location += diff;
}

void CBall::BounceX()
{
	diff.x = -diff.x;
}

void CBall::BounceY()
{
	diff.y = -diff.y;
}

bool CBall::NextMoveInRectX(CRect r)
{
	CRect next = location + diff;
	return (next.left <= r.right) && (next.left >= r.left) &&
			(next.right <= r.right) && (next.right >= r.left);
}

bool CBall::NextMoveInRectY(CRect r)
{
	CRect next = location + diff;
	return (next.top <= r.bottom) && (next.top >= r.top) &&
			(next.bottom <= r.bottom) && (next.bottom >= r.top);
}

bool CBall::NextMoveInRect(CRect r)
{
	return NextMoveInRectX(r) && NextMoveInRectY(r);
}

bool CBall::NextMoveHitRectX(CRect r)
{
	CRect Intersection;
	if (Intersection.IntersectRect(location+diff,r))
		return (Intersection.Width() < Intersection.Height());
	return FALSE;
}

bool CBall::NextMoveHitRectY(CRect r)
{
	CRect Intersection;
	if (Intersection.IntersectRect(location+diff,r))
		return (Intersection.Width() >= Intersection.Height());
	return FALSE;
}

bool CBall::NextMoveHitRect(CRect r)
{
	CRect Intersection;
	return (Intersection.IntersectRect(location+diff,r) != 0);
}

/////////////////////////////////////////////////////////////////////////////
// CAnimation: ѰʵeO
/////////////////////////////////////////////////////////////////////////////

CAnimation::CAnimation(int c)
: DELAY_COUNT(c)
{
	delay_counter = DELAY_COUNT;
}

void CAnimation::AddBitmap(int IDB_BITMAP, int x, int y, COLORREF colorkey) 
{
	CMovingBitmap add_bmp;
	add_bmp.LoadBitmap(IDB_BITMAP, x, y, colorkey);
	bmp.insert(bmp.end(), add_bmp);
	bmp_iter = bmp.begin();
}

CRect CAnimation::GetLocation()
{
	return bmp_iter->GetLocation();
}

int CAnimation::Height()
{
	return bmp_iter->Height();
}

int CAnimation::Left()
{
	return bmp_iter->Left();
}

void CAnimation::SetTopLeft(int nx, int ny)
{
	x = nx, y = ny;
	bmp_iter->SetTopLeft(x, y);
}

void CAnimation::ShowBitmap()
{
	if (--delay_counter <= 0)  {
		delay_counter = DELAY_COUNT;
		bmp_iter++;
		if (bmp_iter == bmp.end())
			bmp_iter = bmp.begin();
	}
	bmp_iter->SetTopLeft(x,y);
	bmp_iter->ShowBitmap();
}

int CAnimation::Top()
{
	return bmp_iter->Top();
}

int CAnimation::Width()
{
	return bmp_iter->Width();
}

/////////////////////////////////////////////////////////////////////////////
// CGame: Game Class
// oclassOC֤ߡAC@ӵ{nJӬ
/////////////////////////////////////////////////////////////////////////////

CGame game;

CGame::CGame()
: NUMBALLS(30) // wyӼ
{
	running = false;
	ball = new CBall [NUMBALLS];
}

CGame::~CGame()
{
	delete [] ball;
}

bool CGame::IsRunning()
{
	return running;
}

void CGame::Move()
{
	//
	// ʭIϪy
	//
	if (background.Top() > SIZE_Y)
		background.SetTopLeft(60 ,-background.Height());
	background.SetTopLeft(background.Left(),background.Top()+1);
	//
	// ʲy
	//
	CRect ClientRect(0,0,SIZE_X,SIZE_Y);
	for (int i=0; i < NUMBALLS; i++) {		// BzC@y
		int soundID = -1;					// ]wSn
		if (!ball[i].NextMoveInRectX(ClientRect))
			ball[i].BounceX();				// y x bϼu
		if (!ball[i].NextMoveInRectY(ClientRect))
			ball[i].BounceY();				// y y bϼu
		if (ball[i].NextMoveHitRectX(racket.GetLocation())) {
			ball[i].BounceX();				// ylA x bϼu
			soundID = AUDIO_CHORD;			// ws1n
		}
		if (ball[i].NextMoveHitRectY(racket.GetLocation())) {
			ball[i].BounceY();				// ylA y bϼu
			soundID = AUDIO_DING;			// ws0n
		}
		if (ball[i].NextMoveInRect(ClientRect) && !ball[i].NextMoveHitRect(racket.GetLocation())) {
			ball[i].Move();
			if (soundID >= 0)
				CAudio::Instance()->Play(soundID);	// wn
		}
	}
	//
	// ʩl
	//
	if (racket.NextMoveInRect(ClientRect)) {
		bool hit = false;					// ˴lʮɷ||y
		for (int i = 0; i < NUMBALLS; i++)
			if (racket.NextMoveHitRect(ball[i].GetLocation())) {
				hit = true;
				break;
			}
		if (!hit)							// Y|yAlyФ~ಾ
			racket.Move();
	}
	//
	//  ʰʵe
	//
	if (animation.Left() > SIZE_X)
		animation.SetTopLeft(-animation.Width(), SIZE_Y/2);
	animation.SetTopLeft(animation.Left()+1,animation.Top());
}

void CGame::OnDraw()
{
	CDDraw::BltBackColor(RGB(0,0,0));	// N Back Plain 
	Show();								// ܹCCӤ
	CDC *pDC = CDDraw::GetBackCDC();	// o Back Plain  CDC 
	CFont f,*fp;
	f.CreatePointFont(160,"з");	//  font f; 16016 pointr
	fp=pDC->SelectObject(&f);			//  font f
	pDC->SetBkColor(RGB(0,0,0));
	pDC->SetTextColor(RGB(255,255,0));
	pDC->TextOut(0,0,"File->New}lAϥηƹlC");
	pDC->TextOut(0,40,"ϥCtrl-FùҦC");
	pDC->SelectObject(fp);				//  font f (dUn|F)
	CDDraw::ReleaseBackCDC();			//  Back Plain  CDC
	CDDraw::BltBackToPrimary();			// N Back Plain Kù
}

void CGame::OnFileNew()
{
	running = true;										// ҰOnIdle()
	for (int i = 0; i < NUMBALLS; i++) {				// ]wy_ly
		ball[i].SetTopLeft(rand()%500, rand()%200);
		ball[i].SetDiffX(rand()%2+2);
		ball[i].SetDiffY(rand()%2+2);
	}
	background.SetTopLeft(60,0);						// ]wI_ly
	animation.SetTopLeft(-animation.Width(), SIZE_Y/2);	// ]wʵe_ly
	CAudio::Instance()->Play(AUDIO_LAKE, true);			//  WAVE
	CAudio::Instance()->Play(AUDIO_CANYON, true);		//  MIDI
	CSpecialEffect::SetCurrentTime();
}
	
void  CGame::OnFilePause()
{
	running = !running;
}

bool CGame::OnIdle()  // ק\ɳ̦nnקOnIdle()AקMove()Show()n
{
	//
	// CO_Ȱ
	//
	if (!running)
		return false;
	//
	// HUOCDj
	//
	CDDraw::BltBackColor(RGB(0,0,0));	// N Back Plain 
	Move();								// ʹCCӤ
	Show();								// ܹCCӤ
	CDDraw::BltBackToPrimary();			// N Back Plain Kù
	//
	// HU{Ci檺tסA`NƶG
	// 1. Build->Start Debug->GoiH˵C@jᱼɶAOɶtC
	// 2. qW}OnIdle()ܦAɶw33msAiRAɶiCtC
	//
	TRACE("Ellipse time for the %d th cycle=%d \n", CSpecialEffect::GetCurrentTimeCount(),CSpecialEffect::GetEllipseTime());
	CSpecialEffect::DelayFromSetCurrentTime(33);
	CSpecialEffect::SetCurrentTime();	// ]w}OnIdle()ɶ
	return true;
}

void CGame::OnInitialUpdate()	// OnInitialUpdate() ub{@}lɰ@
{
	CDDraw::Init(SIZE_X, SIZE_Y);							// ]wCѪR
	//
	// ҦϧΤnӦbo̸J(ӥBunJ@NnFAKoOCPUɶ)
	//
	srand((unsigned)time(NULL));							// Ұʶü
	for (int i = 0; i < NUMBALLS; i++)	
		ball[i].LoadBitmap(IDB_BALL,-100,-100,RGB(0,0,0));	// JiӲyϧΡA(-100,-100)ݤ쪺a
	const int rx = SIZE_X /2;
	const int ry = SIZE_Y-SIZE_Y/4;
	racket.LoadBitmap(IDB_RACKET,rx,ry,RGB(255,255,255));	// Jlϧ
	background.LoadBitmap(IDB_BACKGROUND,-1000,-1000);		// JIϧΡA(-1000,-1000)ݤ쪺a
	corner.LoadBitmap(IDB_CORNER,0,0);						// Jϧ
	corner.ShowBitmap(background);							// NcornerKbackground
	for (i = IDB_BALL1; i <= IDB_BALL4; i++)				// Jʵe(4iϧκc)
		animation.AddBitmap(i,-100,-100,RGB(0,0,0));
	if (!CAudio::Instance()->Open()) {							// }ҭĤ
		AfxMessageBox("Audio Interface Failed (muted)");		// LĤ
		return;
	}
	CAudio::Instance()->Load(AUDIO_DING,  "sounds\\ding.wav");	// Js0nding.wav
	CAudio::Instance()->Load(AUDIO_CHORD, "sounds\\chord.wav");	// Js1nchord.wav
	CAudio::Instance()->Load(AUDIO_LAKE,  "sounds\\lake1.wav");	// Js2nlake1.wav
	CAudio::Instance()->Load(AUDIO_CANYON,"sounds\\canyon.mid");// Js3ncanyon.mid
}

void CGame::OnKillFocus()
{
	running = false;
	CAudio::Instance()->Stop(AUDIO_LAKE);		// WAVE
	CAudio::Instance()->Stop(AUDIO_CANYON);		// MIDI
}

void CGame::OnLButtonDown(UINT nFlags, CPoint point)
{
	racket.SetDiffX(-3);
}

void CGame::OnRButtonDown(UINT nFlags, CPoint point)
{
	racket.SetDiffX(3);
}

void CGame::OnLButtonUp(UINT nFlags, CPoint point)
{
	racket.SetDiffX(0);
}

void CGame::OnRButtonUp(UINT nFlags, CPoint point)
{
	racket.SetDiffX(0);
}

void CGame::Show()
{
	background.ShowBitmap();			// KWIϦ Back Plain
	for (int i=0; i < NUMBALLS; i++)
		ball[i].ShowBitmap();			// KWiy Back Plain
	racket.ShowBitmap();				// KWl Back Plain
	animation.ShowBitmap();				// KWʵe
	//
	//  KWWΥkU
	//
	corner.SetTopLeft(0,0);
	corner.ShowBitmap();
	corner.SetTopLeft(SIZE_X-corner.Width(), SIZE_Y-corner.Height());
	corner.ShowBitmap();
	//
	// NcornerKbackground
	//
	corner.SetTopLeft(0,0);
	corner.ShowBitmap(background);
}